/* This file is a Node.js module that deals with verifying user
 * logins and sessions, where the user information and session
 * information are stored in a SQLite database. It exports the
 * following functions:
 *
 * db_config.configure(options)
 *   Changes the configuration based on the key-value pairs
 *   found in 'options'. This allows configuration of the names
 *   of the tables and columns used. See the 'var config' line
 *   for the different keys that can be used.
 *
 * db_config.register(db, loginId, password, onRegistered)
 *   Adds a new user into the database, with the specified login
 *   ID and unhashed password. Upon completion, the function
 *   calls the onRegistered callback function, passing it three
 *   parameters: an error message (null on success), the ID
 *   allocated to the user, and the ID allocated to the session.
 *
 * db_config.login(db, loginId, submittedPassword, onLogin)
 *   Confirms that the given password matches the hashed password
 *   associated with the given login ID.  Upon completion, the function
 *   calls the onLogin callback function, passing it three
 *   parameters: an error message (null on success), the ID
 *   of the user, and the ID allocated to the session.
 *
 * db_config.retrieveSession(db, sessionId, onRetrieved)
 *   Confirms that the given session ID is valid. Upon completion,
 *   the function calls the onRetrieved callback function,
 *   passing it two parameters: an error message (null on
 *   success) and the user ID of the user with that session.
 *
 * db_config.logout(db, sessionId, onLogout)
 *   Terminates the session with the given ID. Upon completion,
 *   the function calls the onLogout callback function,
 *   passing it an error message (null on success).
 */
var bcrypt = require('bcrypt-nodejs');
var async = require('async');
var crypto = require('crypto');

var config = {
	users: 'users',
	userId: 'userid',
	userLogin: 'login',
	userPassword: 'password',
	sessions: 'sessions',
	sessId: 'sessionid',
	sessUserId: 'userid',
	sessExpires: 'expires',
	expireMinutes: 180
};

//
// private functions
//

function build(template) {
	var result = template.replace(/\{([a-zA-Z]+)\}/g, function (match, key) {
		return config[key];
	});
	console.log(result);
	return result;
}

// onCreated called as onCreated(err, sessionId)
function createSession(db, userId, onCreated) {
	var sessionId;
	var xnInProgress = false;

	async.waterfall([
		function (next) {
			db.run("BEGIN TRANSACTION", next);
		},
		function (next) {
			xnInProgress = true;
			generateRandomUniqueInt(db, config.sessions, config.sessId,
				next);
		},
		function (foundSessionId, next) {
			sessionId = foundSessionId;
			db.run(build(
				"INSERT INTO {sessions} " +
				"({sessId}, {sessUserId}, {sessExpires}) " +
				"VALUES (?, ?, date('now', '+{expiresMinutes} minutes'))"),
				[sessionId, userId], next);
		},
		function (next) {
			db.run("COMMIT TRANSACTION", next);
		},
		function (next) {
			xnInProgress = false;
			onCreated(null, sessionId);
		}
	], function (err) {
		var after = function (err2) {
				if (err !== null) {
					onCreated(err, null);
				} else if (err2 !== null) {
					onCreated('Could not abort transaction', null);
				} else {
					onCreated(null, sessionId);
				}
			};
		if (xnInProgress) {
			db.run("ABORT TRANSACTION", after);
		} else if (err !== null) {
			after(null);
		}
	});
}

function generateRandomUniqueInt(db, tableName, fieldName, onFound) {
	var candidates = [];
	for (var i = 0; i < 5; i += 1) {
		var buf = crypto.randomBytes(4);
		var n = (buf[0] << 23) ^ (buf[1] << 16) ^ (buf[2] << 8) ^ buf[3];
		candidates.push(n);
	}

	db.all("SELECT " + fieldName + " AS id FROM " + tableName +
		" WHERE " + fieldName + " IN (?, ?, ?, ?, ?)",
		candidates, function (err, found) {
			if (err !== null) {
				onFound('SQL error generating new ID', null);
			} else if (found.length === 0) {
				onFound(null, candidates[0]);
			} else {
				for (var i = 0; i < candidates.length; i += 1) {
					var c = candidates[i];
					if (!found.some(function (r) { return r.id === c; })) {
						onFound(null, c);
						return;
					}
				}
				onFound('No session ID is available.', null);
			}
		});
}

//
// exported functions
//

function configure(options) {
	for (var opt in options) {
		if (options.hasOwnProperty(opt)) {
			configOptions[opt] = options[opt];
		}
	}
}

// onRegistered called as onRegistered(err, userId, sessionId)
function register(db, loginId, password, onRegistered) {
	var xnInProgress = false;
	var passwordHash;
	var userId;
	async.waterfall([
		function (next) {
			bcrypt.genSalt(10, next);
		},
		function (salt, next) {
			bcrypt.hash(password, salt, null, next);
		},
		function (hash, next) {
			passwordHash = hash;
			db.run("BEGIN TRANSACTION", next);
		},
		function (next) {
			xnInProgress = true;
			db.all(build("SELECT {userLogin} FROM {users} " +
				"WHERE {userLogin} = ?"), [loginId], next);
		},
		function (found, next) {
			if (found.length !== 0) {
				next('Login ID is already being used.', []);
			} else {
				generateRandomUniqueInt(db, config.users,
					config.userId, next);
			}
		},
		function (foundUserId, next) {
			userId = foundUserId;
			db.run(build("INSERT INTO {users} " +
				"({userId}, {userLogin}, {userPassword}) " +
				" VALUES (?, ?, ?)"),
				[userId, loginId, passwordHash], next);
		},
		function (next) {
			db.run("COMMIT TRANSACTION", next);
		},
		function (next) {
			xnInProgress = false;
			createSession(db, userId, next);
		},
		function (sessionId, next) {
			onRegistered(null, userId, sessionId);
		}
	], function (err) {
		var after = function (err2) {
				if (err !== null) {
					onRegistered(err, null, null);
				} else if (err2 !== null) {
					onRegistered('Could not abort transaction', null, null);
				} else {
					onRegistered('Unknown problem in register', null, null);
				}
			};
		if (xnInProgress) {
			db.run("ABORT TRANSACTION", after);
		} else {
			after(null);
		}
	});
}

// onLogin called as onLogin(err, userId, sessionId)
function login(db, loginId, submittedPassword, onLogin) {
	var userId;
	var hashedPassword;
	async.waterfall([
		function (next) {
			db.all(build(
				"SELECT {userId} AS userid, {userPassword} AS passwd " +
				"FROM  {users} " +
				"WHERE {userLogin} = ?"),
				[loginId], next);
		},
		function (found, next) {
			if (found.length === 0) {
				next('Login ID does not exist.', null);
			} else {
				userId = found[0].userid;
				hashedPassword = found[0].passwd;
				bcrypt.compare(submittedPassword, hashedPassword, next);
			}
		},
		function (match, next) {
			if (match) {
				createSession(db, userId, next);
			} else {
				next('Password does not match', null);
			}
		},
		function (sessionId, next) {
			onLogin(null, userId, sessionId);
		}
	], function (err) {
		if (err !== null) {
			onLogin(err, null, null);
		}
	});
}

// onRetrieved called as onRetrieved(err, userId)
function retrieveSession(db, sessionId, onRetrieved) {
	if (!sessionId) {
		next('You must log in to continue', null);
		return;
	}

	var sql = build("SELECT {sessUserId} AS userid, " +
		"       date('now') > {sessExpires} AS expired " +
		"FROM  {sessions} " +
		"WHERE {sessId} = ? " +
		"ORDER BY {sessExpires} DESC " +
		"LIMIT 1");
	db.all(sql, [sessionId],
		function (err, rows) {
			if (err !== null) {
				console.log('SQL error in retrieveSession', err);
				console.log(sql);
				onRetrieved('SQL error retrieving session', null);
			} else if (rows.length == 0) {
				onRetrieved('You must log in to continue [unknown session]', null);
			} else if (rows[0].expired) {
				onRetrieved('Your session has expired', null);
			} else {
				onRetrieved(null, rows[0].userid);
			}
		});
}

// onLogout called as onLogout(err)
function logout(db, sessionId, onLogout) {
	if (sessionId) {
		db.run(build(
			"DELETE FROM {sessions} " +
			"WHERE {sessId} = ? OR {sessExpires} < date('now', '-1 days')"),
			[sessionId], onLogout);
	} else {
		onLogout(null);
	}
}

exports.configure = configure;
exports.register = register;
exports.login = login;
exports.retrieveSession = retrieveSession;
exports.logout = logout;
