/* This module deals with dynamically compiling a file of Dust.js
 * templates into JavaScript code, ready to be sent to a browser.
 * It continuously monitors when the file of templates has been
 * updated, so that when the file is modified, any later request
 * by the browser for the templates should re-compile the file.
 *
 * The template file can have a number of templates, each given
 * a different label. Templates are marked starting with {@LABEL}
 * and ending with {/LABEL}, where LABEL is the label to be
 * associated with the template.
 */
var dust = require('dustjs-linkedin');
var fs = require('fs');

function compileDustFile(filename, receiver) {
	var fin = fs.createReadStream(filename);
	var finText = '';
	fin.on('data', function (chunk) {
		finText += chunk;
	});
	fin.on('end', function () {
		finText = finText.trim();
		var fullResult = '';
		var reStart = /\s*\{@(\w+)}\s*/g;
		console.log('compiling templates');
		var match = reStart.exec(finText);
		var curPos = 0;
		while (match !== null) {
			if (match.index !== curPos) {
				console.log('  unrecognized characters at', curPos);
			}
			curPos = match.index + match[0].length;
			var templateName = match[1];
			var endMarker = '{/' + templateName + '}';
			var endIndex = finText.indexOf(endMarker, curPos);
			if (endIndex < 0) {
				console.log('  end of template "' + templateName + '" not found');
			} else {
				var templateText = finText.substring(curPos, endIndex);
				curPos = endIndex + endMarker.length;

				console.log('  compiling "' + templateName + '"');
				var compiled = dust.compile(templateText, templateName);
				if (fullResult.length === 0) {
					fullResult = compiled;
				} else {
					fullResult += '\n' + compiled;
				}
			}
			reStart.lastIndex = curPos;
			match = reStart.exec(finText);
		}
		if (curPos !== finText.length) {
			console.log('  unrecognized characters at', curPos);
		}
			
		receiver(fullResult);
	});
}


function middleware(filename, isModifiable) {
	var contents = '';

	var lastCheck = new Date();
	var lastModify = null;

	return function (req, res, next) {
		if (!isModifiable && lastModify !== null) {
			res.send(contents);
			return;
		}
		var now = new Date();
		var elapse = now - lastCheck;
		if (elapse < 1000 && lastModify !== null) {
			res.send(contents);
			return;
		}
		lastCheck = now;
		fs.stat(filename, function (err, stats) {
			if (err !== null) {
				console.log('could not find dust template', filename);
				res.send('');
				return;
			}
			var oldContents = contents;
			if (lastModify === null || stats.mtime > lastModify) {
				lastModify = stats.mtime;
				compileDustFile(filename, function (compiled) {
					contents = compiled;
					res.send(contents);
				});
			} else {
				res.send(oldContents);
			}
		});
	}
};

exports.middleware = middleware;
