var express = require('express');
var app = express();

var path = require('path');
var async = require('async');
var sqlite3 = require('sqlite3');
var db_session = require('./db_session');
var dyn_dust = require('./dyn_dust');

var static_dir = path.normalize(path.join(__dirname, '..', 'static'));
var templ_core = path.normalize(path.join(__dirname, '..', 'static/core.html'));
var templ_login = path.normalize(path.join(__dirname, '..', 'static/login.html'));

app.use(express.logger());
app.use(express.urlencoded());
app.use('/static', express.static(static_dir));
app.use('/template/core.js', dyn_dust.middleware(templ_core, true));
app.use('/template/login.js', dyn_dust.middleware(templ_login, true));
app.use('/', function (req, res, next) {
	if (req.url == '/') {
		res.sendfile(path.join(static_dir, 'index.html'));
	} else {
		next();
	}
});
app.use(express.cookieParser());
app.use(express.cookieSession({ secret: 'site_secret',
	cookie: { maxAge: 3 * 60 * 60 * 1000 }}));
app.use(function (req, res, next) {
	req.db = new sqlite3.Database('flags.db');
	req.db.on('open', next);
	req.db.on('error', function (err) {
		res.send({ ok: false, message: 'internal error: database lost!' });
		console.log('  database lost!', err);
	});
});
app.use(app.router);
app.use(function (req, res, next) {
	ensureDatabaseClosed(req, next);
});

function sendError(res, message) {
	if (message !== null) {
		var msg = (message === undefined ? 'undefined' :
			(message.message || message));
		res.send({ ok: false, message: msg });
		console.log('  Sending error:', message);
	}
	return undefined;
}

function ensureProvided(given, needed) {
	for (var i = 0; i < needed.length; i++) {
		if (!given.hasOwnProperty(needed[i])) {
			return needed[i];
		}
	}
	return null;
}

function ensureDatabaseClosed(req, next) {
	if (req.db) {
		req.db.close(function (err) {
			delete req.db;
			if (next) {
				next(err);
			}
		});
	} else {
		if (next) {
			next(null);
		}
	}
}

function validateLogin(login) {
	if (login === '') {
		return 'Login ID must not be empty.';
	} else if (!login) {
		return 'Login ID must be provided.';
	} else if (!/^[a-zA-Z0-9_.-]+$/.test(login)) {
		return 'Login ID can contain only letters and digits.';
	} else {
		return null;
	}
}

function validatePassword(password) {
	if (password === '') {
		return 'Password must not be empty.';
	} else if (!password) {
		return 'Password must be provided.';
	} else if (password.length < 4) {
		return 'Password is too short.';
	} else {
		return null;
	}
}

function validateFirstName(name) {
	if (name === '') {
		return 'First name must not be empty.';
	} else if (!name) {
		return 'First name must be provided.';
	} else if (!/^[a-zA-Z'.-]+$/.test(name)) {
		return 'First name must contain only letters.';
	} else {
		return null;
	}
}

function validateLastName(name) {
	if (name === '') {
		return 'Last name must not be empty.';
	} else if (!name) {
		return 'Last name must be provided.';
	} else if (!/^[a-zA-Z'.-]+$/.test(name)) {
		return 'Last name must contain only letters.';
	} else {
		return null;
	}
}

app.post('/register', function (req, res) {
	var missing = ensureProvided(req.body,
		['login', 'password', 'firstname', 'lastname']);
	if (missing) {
		ensureDatabaseClosed(req);
		return sendError(res, 'AJAX /register request needs ' + missing + ' field');
	}

	var login = req.body['login'];
	var password = req.body['password'];
	var firstname = req.body['firstname']
	var lastname = req.body['lastname']

	var err = (validateLogin(login) || validatePassword(password) ||
		validateFirstName(firstname) || validateLastName(lastname));
	if (err) {
		ensureDatabaseClosed(req);
		return sendError(res, err);
	}

	var userId;
	async.waterfall([
		function (next) {
			db_session.register(req.db, login, password, next);
		},
		function (foundUserId, sessionId, next) {
			userId = foundUserId;
			req.session.sessionId = sessionId;

			req.db.run("UPDATE users " +
				"SET firstname = ?, lastname = ? WHERE userid = ?",
				[firstname, lastname, userId], next);
		},
		function (next) {
			ensureDatabaseClosed(req, next);
		},
		function (next) {
			res.send({ ok: true, userid: userId });
		}
	], function (err) {
		ensureDatabaseClosed(req);
		sendError(res, err);
	});
});
app.post('/login', function (req, res) {
	var missing = ensureProvided(req.body, ['login', 'password']);
	if (missing) {
		ensureDatabaseClosed(req);
		return sendError(res, 'AJAX /login request needs ' + missing + ' field');
	}

	var login = req.body['login'];
	var password = req.body['password'];

	var userId;
	async.waterfall([
		function (next) {
			db_session.login(req.db, login, password, next);
		},
		function (foundUserId, sessionId, next) {
			userId = foundUserId;
			req.session.sessionId = sessionId;
			ensureDatabaseClosed(req, next);
		},
		function (next) {
			res.send({ ok: true, userid: userId });
		}
	], function (err) {
		ensureDatabaseClosed(req);
		sendError(res, err);
	});
});
app.post('/logout', function (req, res) {
	var sessionId = req.session.sessionId;
	if (!sessionId) {
		ensureDatabaseClosed(req);
		res.send({ ok: true });
	} else {
		delete req.session.sessionId;
		db_session.logout(req.db, sessionId, function (err) {
			ensureDatabaseClosed(req);
			if (err !== null) {
				sendError(res, err);
			} else {
				res.send({ ok: true });
			}
		});
	}
});
app.post('/order', function (req, res) {
	var missing = ensureProvided(req.body, ['productid', 'price', 'address']);
	if (missing) {
		ensureDatabaseClosed(req);
		return sendError(res, 'AJAX /order request needs ' + missing + ' field');
	}

	var productId = req.body['productid'];
	var price = req.body['price'];
	var address = req.body['address'];

	if (address === '') {
		ensureDatabaseClosed(req);
		return sendError(res, 'Address cannot be empty.');
	} else if (!/^[A-Za-z0-9 .#&():,-]+$/.test(address)) {
		ensureDatabaseClosed(req);
		return sendError(res, 'Address contains invalid characters.');
	}

	var userId;
	var xnStarted = false;
	async.waterfall([
		function (next) {
			db_session.retrieveSession(req.db, req.session.sessionId, next);
		},
		function (foundUserId, next) {
			userId = foundUserId;
			req.db.run("BEGIN TRANSACTION", next);
		},
		function (next) {
			xnStarted = true;
			req.db.all("SELECT inventory, price " +
				"FROM products WHERE productid = ?",
				[productId], next);
		},
		function (found, next) {
			if (found.length !== 1) {
				next('Product with that ID does not exist.');
			} else if (found[0].inventory <= 0) {
				next('Product is no longer available.');
			} else {
				var foundPrice = found[0].price;
				if (found[0].inventory <= 1) {
					foundPrice += 1;
				}
				if (parseFloat(foundPrice) !== parseFloat(price)) {
					console.log(foundPrice, price);
					next('Product\'s price has changed.');
				} else {
					req.db.run("UPDATE products SET inventory = inventory - 1 "
						+ "WHERE productid = ?", [productId], next);
				}
			}
		},
		function (next) {
			req.db.run("INSERT INTO orders (userid, productid, price, address) " +
				"VALUES (?, ?, ?, ?)", [userId, productId, price, address], next);
		},
		function (next) {
			req.db.run("COMMIT TRANSACTION", next);
		},
		function (next) {
			ensureDatabaseClosed(req, next);
		},
		function (next) {
			xnStarted = false;
			res.send({ ok: true });
		}
	], function (err) {
		if (xnStarted) {
			req.db.run("ABORT TRANSACTION", function (err2) {
				ensureDatabaseClosed(req);
				sendError(res, err || err2);
			});
		} else {
			ensureDatabaseClosed(req);
			sendError(res, err);
		}
	});
});
app.get('/search', function (req, res) {
	var missing = ensureProvided(req.query, ['query']);
	if (missing) {
		ensureDatabaseClosed(req);
		return sendError(res, 'AJAX /search request needs ' + missing + ' field');
	}

	req.db.all("SELECT productid, name, price, inventory " +
		"FROM products WHERE name LIKE ? ORDER BY name LIMIT 11",
		['%' + req.query.query + '%'], function (err, rows) {
			ensureDatabaseClosed(req);
			if (err !== null) {
				sendError(res, err);
			} else {
				var more = rows.length > 10;
				results = [];
				for (var i = 0; i < rows.length && i < 10; i++) {
					var iPrice = rows[i].price;
					if (rows[i].inventory <= 1) {
						iPrice += 1;
					}
					results.push({ productid: rows[i].productid,
						name: rows[i].name, price: iPrice,
						avail: rows[i].inventory > 0 });
				}
				res.send({ ok: true, results: results, more: more });
			}
		});
});
app.get('/orders', function (req, res) {
	async.waterfall([
		function (next) {
			db_session.retrieveSession(req.db, req.session.sessionId, next);
		},
		function (userId, next) {
			req.db.all("SELECT ordertime, name, orders.price AS price, address " +
				"FROM orders LEFT JOIN products ON orders.productid = products.productid " +
				"WHERE userid = ? ORDER BY ordertime DESC", [userId], next);
		},
		function (rows, next) {
			ensureDatabaseClosed(req);
			res.send({ ok: true, results: rows });
		}
	], function (err) {
		ensureDatabaseClosed(req);
		sendError(res, err);
	});
});

app.listen(8000);
console.log("Server ready. Access via http://localhost:8000");
